from .bmath.unit import *
from .drag import *


cdef class Projectile:
    cdef _bullet_diameter
    cdef _bullet_length
    cdef _ballistic_coefficient
    cdef _has_dimensions
    cdef _weight

    def __init__(self, ballistic_coefficient: BallisticCoefficient, weight: Weight):
        """
        projectile description with dimensions
        :param ballistic_coefficient: BallisticCoefficient instance
        :param weight: unit.Weight instance
        """
        self._ballistic_coefficient = ballistic_coefficient
        self._has_dimensions = False
        self._weight = weight

    cpdef ballistic_coefficient(self):
        return self._ballistic_coefficient

    cpdef bullet_weight(self):
        return self._weight

    cpdef bullet_diameter(self):
        return self._bullet_diameter

    cpdef bullet_length(self):
        return self._bullet_length

    cpdef has_dimensions(self):
        return self._has_dimensions


cdef class ProjectileWithDimensions(Projectile):

    def __init__(self, ballistic_coefficient: BallisticCoefficient,
                 bullet_diameter: Distance,
                 bullet_length: Distance,
                 weight: Weight):
        super(ProjectileWithDimensions, self).__init__(ballistic_coefficient, weight)
        self._has_dimensions = True
        self._bullet_diameter = bullet_diameter
        self._bullet_length = bullet_length

cdef class Ammunition:
    cdef _projectile
    cdef _muzzle_velocity

    def __init__(self, bullet: Projectile, muzzle_velocity: Velocity):
        self._projectile = bullet
        self._muzzle_velocity = muzzle_velocity

    cpdef bullet(self):
        return self._projectile

    cpdef muzzle_velocity(self):
        return self._muzzle_velocity



from libc.math cimport fabs, pow, sin, cos, log10, floor, atan
from .bmath.unit import *
from .bmath.vector import Vector
from .projectile import Ammunition
from .weapon import Weapon, TwistLeft
from .atmosphere import Atmosphere
from .shot_parameters import ShotParameters
from .wind import WindInfo
from .trajectory_data import TrajectoryData, Timespan

cdef double cZeroFindingAccuracy = 0.000005
cdef double cMinimumVelocity = 50.0
cdef double cMaximumDrop = -15000
cdef int cMaxIterations = 10
cdef double cGravityConstant = -32.17405

cdef class TrajectoryCalculator:
    cdef _maximum_calculator_step_size

    def __init__(self):
        self._maximum_calculator_step_size = Distance(1, DistanceFoot)

    cpdef maximum_calculator_step_size(self):
        return self._maximum_calculator_step_size

    cpdef set_maximum_calculator_step_size(self, value: Distance):
        self._maximum_calculator_step_size = value

    cdef double get_calculation_step(self, double step):
        cdef step_order, maximum_order
        step = step / 2
        cdef double maximum_step = self._maximum_calculator_step_size.get_in(DistanceFoot)

        if step > maximum_step:
            step_order = int(floor(log10(step)))
            maximum_order = int(floor(log10(maximum_step)))
            step = step / pow(10, float(step_order - maximum_order + 1))

        return step

    cpdef sight_angle(self, ammunition: Ammunition, weapon: Weapon, atmosphere: Atmosphere):
        cdef double calculation_step, mach, density_factor, muzzle_velocity
        cdef double barrel_azimuth, barrel_elevation
        cdef double velocity, time, zero_distance, maximum_range
        cdef double delta_time, drag, zero_finding_error

        cdef int iterations_count

        cdef gravity_vector, range_vector, velocity_vector, delta_range_vector

        calculation_step = self.get_calculation_step(
            Distance(10, weapon.zero().zero_distance().units()).get_in(DistanceFoot))

        mach = atmosphere.mach().get_in(VelocityFPS)
        density_factor = atmosphere.density_factor()
        muzzle_velocity = ammunition.muzzle_velocity().get_in(VelocityFPS)
        barrel_azimuth = 0.0
        barrel_elevation = 0.0

        iterations_count = 0

        zero_finding_error = cZeroFindingAccuracy * 2
        gravity_vector = Vector(0, cGravityConstant, 0)
        while zero_finding_error > cZeroFindingAccuracy and iterations_count < cMaxIterations:
            velocity = muzzle_velocity
            time = 0.0

            # x - distance towards target,
            # y - drop and
            # z - windage

            range_vector = Vector(
                0.0, -weapon.sight_height().get_in(DistanceFoot), 0.0
            )
            velocity_vector = Vector(
                cos(barrel_elevation) * cos(barrel_azimuth),
                sin(barrel_elevation),
                cos(barrel_elevation) * sin(barrel_azimuth)
            ).multiply_by_const(velocity)

            zero_distance = weapon.zero().zero_distance().get_in(DistanceFoot)
            maximum_range = zero_distance + calculation_step

            while range_vector.x() <= maximum_range:
                if velocity < cMinimumVelocity or range_vector.y() < cMaximumDrop:
                    break

                delta_time = calculation_step / velocity_vector.x()
                velocity = velocity_vector.magnitude()
                drag = density_factor * velocity * ammunition \
                    .bullet() \
                    .ballistic_coefficient() \
                    .drag(velocity / mach)

                velocity_vector = velocity_vector.subtract(
                    (
                        velocity_vector.multiply_by_const(drag).subtract(
                            gravity_vector)
                    ).multiply_by_const(delta_time)
                )

                delta_range_vector = Vector(calculation_step,
                                            velocity_vector.y() * delta_time,
                                            velocity_vector.z() * delta_time)
                range_vector = range_vector.add(delta_range_vector)
                velocity = velocity_vector.magnitude()
                time = time + delta_range_vector.magnitude() / velocity
                if fabs(range_vector.x() - zero_distance) < 0.5 * calculation_step:
                    zero_finding_error = fabs(range_vector.y())
                    barrel_elevation = barrel_elevation - range_vector.y() / range_vector.x()
                    break

                iterations_count += 1
        return Angular(barrel_elevation, AngularRadian)

    cpdef trajectory(self, ammunition: Ammunition, weapon: Weapon, atmosphere: Atmosphere,
                     shot_info: ShotParameters, wind_info: list[WindInfo]):
        cdef double range_to, step, calculation_step, bullet_weight, stability_coefficient
        cdef double barrel_azimuth, barrel_elevation, alt0, density_factor, mach
        cdef double next_wind_range, time, muzzle_velocity, velocity, windage, delta_time, drag
        cdef double maximum_range, next_range_distance, twist_coefficient
        cdef int current_item, ranges_length, current_wind, len_wind_info
        cdef calculate_drift, ranges, wind_vector
        cdef windage_adjustment, velocity_adjusted, delta_range_vector
        cdef gravity_vector, drop_adjustment, range_vector, velocity_vector

        range_to = shot_info.maximum_distance().get_in(DistanceFoot)
        step = shot_info.step().get_in(DistanceFoot)

        calculation_step = self.get_calculation_step(step)

        bullet_weight = ammunition.bullet().bullet_weight().get_in(WeightGrain)

        stability_coefficient = 1.0

        calculate_drift = False

        if weapon.has_twist and ammunition.bullet().has_dimensions():
            stability_coefficient = calculate_stability_coefficient(
                ammunition, weapon, atmosphere
            )
            calculate_drift = True

        ranges_length = int(floor(range_to / step)) + 1
        ranges = []

        barrel_azimuth = .0
        barrel_elevation = shot_info.sight_angle().get_in(AngularRadian)
        barrel_elevation = barrel_elevation + shot_info.shot_angle().get_in(AngularRadian)
        alt0 = atmosphere.altitude().get_in(DistanceFoot)

        # Never used in upstream, uncomment on need
        # density_factor, mach = atmosphere.get_density_factor_and_mach_for_altitude(alt0)

        current_wind = 0
        next_wind_range = 1e7

        len_wind_info = int(len(wind_info))

        if len_wind_info < 1:
            wind_vector = Vector(0, 0, 0)
        else:
            if len(wind_info) > 1:
                next_wind_range = wind_info[0].until_distance().get_in(DistanceFoot)
            wind_vector = wind_to_vector(shot_info, wind_info[0])

        muzzle_velocity = ammunition.muzzle_velocity().get_in(VelocityFPS)
        gravity_vector = Vector(0, cGravityConstant, 0)
        velocity = muzzle_velocity
        time = .0

        # x - distance towards target,
        # y - drop and
        # z - windage

        range_vector = Vector(.0, -weapon.sight_height().get_in(DistanceFoot), 0)
        velocity_vector = Vector(cos(barrel_elevation) * cos(barrel_azimuth), sin(barrel_elevation),
                                 cos(barrel_elevation) * sin(barrel_azimuth)).multiply_by_const(velocity)
        current_item = 0

        maximum_range = range_to
        next_range_distance = 0

        twist_coefficient = .0

        if calculate_drift:
            if weapon.twist().direction() == TwistLeft:
                twist_coefficient = 1
            else:
                twist_coefficient = -1

        while range_vector.x() <= maximum_range + calculation_step:

            if velocity < cMinimumVelocity or range_vector.y() < cMaximumDrop:
                break

            density_factor, mach = atmosphere.get_density_factor_and_mach_for_altitude(alt0 + range_vector.y())

            if range_vector.x() >= next_wind_range:
                current_wind += 1
                wind_vector = wind_to_vector(shot_info, wind_info[current_wind])

                if current_wind == len_wind_info - 1:
                    next_wind_range = 1e7
                else:
                    next_wind_range = wind_info[current_wind].until_distance().get_in(DistanceFoot)

            if range_vector.x() >= next_range_distance:
                windage = range_vector.z()
                if calculate_drift:
                    windage += (1.25 * (stability_coefficient + 1.2) * pow(time, 1.83) * twist_coefficient) / 12

                drop_adjustment = get_correction(range_vector.x(), range_vector.y())
                windage_adjustment = get_correction(range_vector.x(), windage)

                ranges.append(TrajectoryData(
                    time=Timespan(time),
                    travel_distance=Distance(range_vector.x(), DistanceFoot),
                    drop=Distance(range_vector.y(), DistanceFoot),
                    drop_adjustment=Angular(drop_adjustment, AngularRadian) if drop_adjustment else None,
                    windage=Distance(windage, DistanceFoot),
                    windage_adjustment=Angular(windage_adjustment, AngularRadian) if windage_adjustment else None,
                    velocity=Velocity(velocity, VelocityFPS),
                    mach=velocity / mach,
                    energy=Energy(calculate_energy(bullet_weight, velocity),
                                  EnergyFootPound),
                    optimal_game_weight=Weight(calculate_ogv(bullet_weight, velocity),
                                               WeightPound))
                )

                next_range_distance += step
                current_item += 1
                if current_item == ranges_length:
                    break

            delta_time = calculation_step / velocity_vector.x()
            velocity_adjusted = velocity_vector.subtract(wind_vector)
            velocity = velocity_adjusted.magnitude()

            drag = density_factor * velocity * ammunition \
                .bullet() \
                .ballistic_coefficient() \
                .drag(velocity / mach)

            velocity_vector = velocity_vector.subtract(
                (
                    velocity_adjusted.multiply_by_const(drag).subtract(
                        gravity_vector)
                ).multiply_by_const(delta_time)
            )
            delta_range_vector = Vector(calculation_step,
                                        velocity_vector.y() * delta_time,
                                        velocity_vector.z() * delta_time)
            range_vector = range_vector.add(delta_range_vector)
            velocity = velocity_vector.magnitude()
            time = time + delta_range_vector.magnitude() / velocity

        return ranges


cdef double calculate_stability_coefficient(ammunition_info, rifle_info, atmosphere):
    cdef double weight = ammunition_info.bullet().bullet_weight().get_in(WeightGrain)
    cdef double diameter = ammunition_info.bullet().bullet_diameter().get_in(DistanceInch)
    cdef double twist = rifle_info.twist().twist().get_in(DistanceInch) / diameter
    cdef double length = ammunition_info.bullet().bullet_length().get_in(DistanceInch) / diameter
    cdef double sd = 30 * weight / (pow(twist, 2) * pow(diameter, 3) * length * (1 + pow(length, 2)))
    cdef double mv = ammunition_info.muzzle_velocity().get_in(VelocityFPS)
    cdef double fv = pow(mv / 2800, 1.0 / 3.0)
    cdef double ft = atmosphere.temperature().get_in(TemperatureFahrenheit)
    cdef double pt = atmosphere.pressure().get_in(PressureInHg)
    cdef double ftp = ((ft + 460) / (59 + 460)) * (29.92 / pt)

    return sd * fv * ftp

cdef wind_to_vector(shot, wind):
    cdef double sight_cosine = cos(shot.sight_angle().get_in(AngularRadian))
    cdef double sight_sine = sin(shot.sight_angle().get_in(AngularRadian))
    cdef double cant_cosine = cos(shot.cant_angle().get_in(AngularRadian))
    cdef double cant_sine = sin(shot.cant_angle().get_in(AngularRadian))
    cdef double range_velocity = wind.velocity().get_in(VelocityFPS) * cos(wind.direction().get_in(AngularRadian))
    cdef double cross_component = wind.velocity().get_in(VelocityFPS) * sin(wind.direction().get_in(AngularRadian))
    cdef double range_factor = -range_velocity * sight_sine
    return Vector(range_velocity * sight_cosine,
                  range_factor * cant_cosine + cross_component * cant_sine,
                  cross_component * cant_cosine - range_factor * cant_sine)

cdef get_correction(double distance, double offset):
    if distance != 0:
        return atan(offset / distance)
    return None

cdef double calculate_energy(double bullet_weight, double velocity):
    return bullet_weight * pow(velocity, 2) / 450400

cdef double calculate_ogv(double bullet_weight, double velocity):
    return pow(bullet_weight, 2) * pow(velocity, 3) * 1.5e-12



WeightGrain = 70
WeightOunce = 71
WeightGram = 72
WeightPound = 73
WeightKilogram = 74
WeightNewton = 75


cdef class Weight:
    cdef double _value
    cdef int _default_units
    cdef __name__

    def __init__(self, value: double, units: int):
        self.__name__ = 'Weight'
        self._value = self.to_default(value, units)
        self._default_units = units

    cdef double to_default(self, value: double, units: int):
        if units == WeightGrain:
            return value
        elif units == WeightGram:
            return value * 15.4323584
        elif units == WeightKilogram:
            return value * 15432.3584
        elif units == WeightNewton:
            return value * 151339.73750336
        elif units == WeightPound:
            return value / 0.000142857143
        elif units == WeightOunce:
            return value * 437.5
        else:
            raise KeyError(f'{self.__name__}: unit {units} is not supported')

    cdef double from_default(self, value: double, units: int):
        if units == WeightGrain:
            return value
        elif units == WeightGram:
            return value / 15.4323584
        elif units == WeightKilogram:
            return value / 15432.3584
        elif units == WeightNewton:
            return value / 151339.73750336
        elif units == WeightPound:
            return value * 0.000142857143
        elif units == WeightOunce:
            return value / 437.5
        else:
            raise KeyError(f'KeyError: {self.__name__}: unit {units} is not supported')

    cpdef double value(self, units: int):
        return self.from_default(self._value, units)

    cpdef Weight convert(self, units: int):
        cdef double value = self.get_in(units)
        return Weight(value, units)

    cpdef double get_in(self, units: int):
        return self.from_default(self._value, units)

    def __str__(self):
        return self.string()

    cdef string(self):
        cdef name
        cdef int accuracy
        cdef int default = self._default_units
        cdef double v = self.from_default(self._value, default)
        if default == WeightGrain:
            name = 'gr'
            accuracy = 0
        elif default == WeightGram:
            name = 'g'
            accuracy = 1
        elif default == WeightKilogram:
            name = 'kg'
            accuracy = 3
        elif default == WeightNewton:
            name = 'N'
            accuracy = 3
        elif default == WeightPound:
            name = 'lb'
            accuracy = 3
        elif default == WeightOunce:
            name = 'oz'
            accuracy = 1
        else:
            name = '?'
            accuracy = 6

        return f'{round(v, accuracy)} {name}'

    cpdef int units(self):
        return self._default_units



__author__ = "o-murphy"
__copyright__ = ("",)

__credits__ = ["O-Murphy"]
__version__ = "1.0.11"

__all__ = ['profile', 'atmosphere', 'bmath', 'drag', 'drag_tables', 'projectile', 'shot_parameters',
           'trajectory_calculator', 'trajectory_data', 'weapon', 'wind', 'multiple_bc']



from .bmath.unit import *


cdef class WindInfo:
    cdef _until_distance
    cdef _velocity
    cdef _direction

    def __init__(self, until_distance: Distance = None,
                 velocity: Velocity = None, direction: Angular = None):
        self._until_distance = until_distance
        self._velocity = velocity
        self._direction = direction

    def __str__(self):
        return f'Until distance: {self._until_distance}, Velocity: {self._velocity}, Direction: {self._direction}'

    cdef string(self):
        return f'Until distance: {self._until_distance}, Velocity: {self._velocity}, Direction: {self._direction}'

    cpdef until_distance(self):
        return self._until_distance

    cpdef velocity(self):
        return self._velocity

    cpdef direction(self):
        return self._direction

cpdef create_no_wind():
    w = WindInfo()
    return [w]

cpdef create_only_wind_info(wind_velocity: Velocity, direction: Angular):
    cdef until_distance, w
    until_distance = Distance(9999, DistanceKilometer)
    w = [WindInfo(until_distance, wind_velocity, direction)]
    return w

# cpdef add_wind_info(until_distance: Distance,
#                     velocity: Velocity, direction: Angular):
#     return WindInfo(until_distance, velocity, direction)
#
# cdef Wind

def create_wind_info(*winds: 'WindInfo') -> list['WindInfo']:
    return list(winds)



VelocityMPS = 60
VelocityKMH = 61
VelocityFPS = 62
VelocityMPH = 63
VelocityKT = 64


cdef class Velocity:
    cdef double _value
    cdef int _default_units
    cdef __name__

    def __init__(self, value: double, units: int):
        self.__name__ = 'Energy'
        self._value = self.to_default(value, units)
        self._default_units = units

    cdef double to_default(self, value: double, units: int):
        if units == VelocityMPS:
            return value
        elif units == VelocityKMH:
            return value / 3.6
        elif units == VelocityFPS:
            return value / 3.2808399
        elif units == VelocityMPH:
            return value / 2.23693629
        elif units == VelocityKT:
            return value / 1.94384449
        else:
            raise KeyError(f'{self.__name__}: unit {units} is not supported')

    cdef double from_default(self, value: double, units: int):
        if units == VelocityMPS:
            return value
        elif units == VelocityKMH:
            return value * 3.6
        elif units == VelocityFPS:
            return value * 3.2808399
        elif units == VelocityMPH:
            return value * 2.23693629
        elif units == VelocityKT:
            return value * 1.94384449
        else:
            raise KeyError(f'KeyError: {self.__name__}: unit {units} is not supported')

    cpdef double value(self, units: int):
        return self.from_default(self._value, units)

    cpdef Velocity convert(self, units: int):
        cdef double value = self.get_in(units)
        return Velocity(value, units)

    cpdef double get_in(self, units: int):
        return self.from_default(self._value, units)

    def __str__(self):
        return self.string()

    cdef string(self):
        cdef name
        cdef int accuracy
        cdef int default = self._default_units
        cdef double v = self.from_default(self._value, default)
        if default == VelocityMPS:
            name = "m/s"
            accuracy = 0
        elif default == VelocityKMH:
            name = "km/h"
            accuracy = 1
        elif default == VelocityFPS:
            name = "ft/s"
            accuracy = 1
        elif default == VelocityMPH:
            name = "mph"
            accuracy = 1
        elif default == VelocityKT:
            name = "kt"
            accuracy = 1
        else:
            name = '?'
            accuracy = 6
        return f'{round(v, accuracy)} {name}'

    cpdef int units(self):
        return self._default_units



import timeit
import unittest
import pyximport
import math
pyximport.install()
from py_ballisticcalc.bmath.vector import Vector


class TestVector(unittest.TestCase):

    def test_create(self):
        v = Vector(1, 2, 3)
        mag = v.magnitude()
        mc = v.multiply_by_const(10)

        b = Vector(1, 2, 3)
        mv = v.multiply_by_vector(b)
        a = v.add(b)
        s = v.subtract(b)
        neg = v.negate()
        norm = v.normalize()
        # print('\n'.join([str(i) for i in [mag, mc, mv, a, s, neg, norm]]))
        x = v.x

    # @unittest.SkipTest
    def test_time(self):
        t = timeit.timeit(self.test_create, number=50000)
        print(t)


class TestVectorCreation(unittest.TestCase):

    def test_create(self):
        v = Vector(1, 2, 3)

        self.assertTrue(v.x == 1 and v.y == 2 and v.z == 3, "Create failed")

        c = v.copy()
        self.assertTrue(c.x == 1 and c.y == 2 and c.z == 3, "Copy failed")


class TestUnary(unittest.TestCase):

    def test_unary(self):
        v1 = Vector(1, 2, 3)

        self.assertFalse(math.fabs(v1.magnitude() - 3.74165738677) > 1e-7, "Magnitude failed")

        v2 = v1.negate()
        self.assertTrue(v2.x() == -1.0 and v2.y() == -2.0 and v2.z() == -3.0, "Negate failed")

        v2 = v1.normalize()
        self.assertFalse(v2.x() > 1.0 or v2.y() > 1.0 or v2.z() > 1.0, "Normalize failed")

        v1 = Vector(0, 0, 0)
        v2 = v1.normalize()

        self.assertFalse(v2.x() != 0.0 or v2.y() != 0.0 or v2.z() != 0.0, "Normalize failed")


class TestBinary(unittest.TestCase):

    def test_binary(self):
        v1 = Vector(1, 2, 3)

        v2 = v1.add(v1.copy())
        self.assertFalse(v2.x() != 2.0 or v2.y() != 4.0 or v2.z() != 6.0, "Add failed")

        v2 = v1.subtract(v2)
        self.assertFalse(v2.x() != -1.0 or v2.y() != -2.0 or v2.z() != -3.0, "Subtract failed")

        self.assertFalse(v1.multiply_by_vector(v1.copy()) != float(1 + 4 + 9), "MultiplyByVector failed")

        v2 = v1.multiply_by_const(3)
        self.assertFalse(v2.x() != 3.0 or v2.y() != 6.0 or v2.z() != 9.0, "MultiplyByConst failed")



import unittest
import pyximport
pyximport.install()
from py_ballisticcalc.bmath.unit import *

import math


def test_back_n_forth(test, value, units):
    u = test.unit_class(value, units)
    v = u.value(units)
    test.assertTrue(
        math.fabs(v - value) < 1e-7
        and math.fabs(v - u.get_in(units) < 1e-7), f'Read back failed for {units}')


class TestAngular(unittest.TestCase):
    def setUp(self) -> None:
        self.unit_class = Angular
        self.unit_list = [
            AngularDegree,
            AngularMOA,
            AngularMRad,
            AngularMil,
            AngularRadian,
            AngularThousand
        ]

    def test_angular(self):
        for u in self.unit_list:
            with self.subTest(unit=u):
                test_back_n_forth(self, 3, u)


class TestDistance(unittest.TestCase):
    def setUp(self) -> None:
        self.unit_class = Distance
        self.unit_list = [
            DistanceCentimeter,
            DistanceFoot,
            DistanceInch,
            DistanceKilometer,
            DistanceLine,
            DistanceMeter,
            DistanceMillimeter,
            DistanceMile,
            DistanceNauticalMile,
            DistanceYard
        ]

    def test_distance(self):
        for u in self.unit_list:
            with self.subTest(unit=u):
                test_back_n_forth(self, 3, u)


class TestEnergy(unittest.TestCase):
    def setUp(self) -> None:
        self.unit_class = Energy
        self.unit_list = [
            EnergyFootPound,
            EnergyJoule
        ]

    def test_energy(self):
        for u in self.unit_list:
            with self.subTest(unit=u):
                test_back_n_forth(self, 3, u)


class TestPressure(unittest.TestCase):
    def setUp(self) -> None:
        self.unit_class = Pressure
        self.unit_list = [
            PressureBar,
            PressureHP,
            PressureMmHg,
            PressureInHg
        ]

    def test_pressure(self):
        for u in self.unit_list:
            with self.subTest(unit=u):
                test_back_n_forth(self, 3, u)


class TestTemperature(unittest.TestCase):
    def setUp(self) -> None:
        self.unit_class = Temperature
        self.unit_list = [
            TemperatureFahrenheit,
            TemperatureKelvin,
            TemperatureCelsius,
            TemperatureRankin
        ]

    def test_temperature(self):
        for u in self.unit_list:
            with self.subTest(unit=u):
                test_back_n_forth(self, 3, u)


class TestVelocity(unittest.TestCase):
    def setUp(self) -> None:
        self.unit_class = Velocity
        self.unit_list = [
            VelocityFPS,
            VelocityKMH,
            VelocityKT,
            VelocityMPH,
            VelocityMPS
        ]

    def test_velocity(self):
        for u in self.unit_list:
            with self.subTest(unit=u):
                test_back_n_forth(self, 3, u)


class TestWeight(unittest.TestCase):
    def setUp(self) -> None:
        self.unit_class = Weight
        self.unit_list = [
            WeightGrain,
            WeightGram,
            WeightKilogram,
            WeightNewton,
            WeightOunce,
            WeightPound
        ]

    def test_weight(self):
        for u in self.unit_list:
            with self.subTest(unit=u):
                test_back_n_forth(self, 3, u)


if __name__ == '__main__':
    unittest.main()



import timeit
from datetime import datetime
import unittest
import pyximport

from math import fabs

pyximport.install()
from py_ballisticcalc.atmosphere import Atmosphere, IcaoAtmosphere
from py_ballisticcalc.drag import BallisticCoefficient, DragTableG1, DragTableG7
from py_ballisticcalc.projectile import Projectile, ProjectileWithDimensions
from py_ballisticcalc.weapon import Ammunition, ZeroInfo, TwistInfo, TwistRight, WeaponWithTwist, Weapon
from py_ballisticcalc.wind import create_only_wind_info
from py_ballisticcalc.shot_parameters import ShotParameters, ShotParametersUnlevel
from py_ballisticcalc.trajectory_calculator import TrajectoryCalculator
from py_ballisticcalc.trajectory_data import TrajectoryData
from py_ballisticcalc.bmath import unit as unit
from py_ballisticcalc.bmath.unit import *
from py_ballisticcalc.profile import Profile as cProfile
from py_ballisticcalc.multiple_bc import MultipleBallisticCoefficient


class TestProfile(unittest.TestCase):
    """
    0.22300000488758087
    -9.000000953674316 0.0
    -0.00047645941958762705 100.0496826171875
    -188.0503692626953 500.03924560546875
    -1475.96826171875 1000.0016479492188
    1.5700003132224083e-05 def init
    0.09897580003598705 def init + make
    0.2844648000318557 max=2500m, step=1m
    0.04717749997507781 max=2500m, step=1m, max_step=5ft
    """

    # @unittest.SkipTest
    def test_profile_bc(self):

        p = cProfile()

        data = p.calculate_trajectory()

        print(data[0].drop().get_in(DistanceCentimeter), data[0].travelled_distance().get_in(DistanceMeter))
        print(data[1].drop().get_in(DistanceCentimeter), data[1].travelled_distance().get_in(DistanceMeter))
        print(data[5].drop().get_in(DistanceCentimeter), data[5].travelled_distance().get_in(DistanceMeter))
        print(data[10].drop().get_in(DistanceCentimeter), data[10].travelled_distance().get_in(DistanceMeter))


        p.calculate_drag_table()
        print(p.dict())

    def test_custom_df(self):
        custom_drag_func = [
            {'A': 0.0, 'B': 0.18}, {'A': 0.4, 'B': 0.178}, {'A': 0.5, 'B': 0.154},
            {'A': 0.6, 'B': 0.129}, {'A': 0.7, 'B': 0.131}, {'A': 0.8, 'B': 0.136},
            {'A': 0.825, 'B': 0.14}, {'A': 0.85, 'B': 0.144}, {'A': 0.875, 'B': 0.153},
            {'A': 0.9, 'B': 0.177}, {'A': 0.925, 'B': 0.226}, {'A': 0.95, 'B': 0.26},
            {'A': 0.975, 'B': 0.349}, {'A': 1.0, 'B': 0.427}, {'A': 1.025, 'B': 0.45},
            {'A': 1.05, 'B': 0.452}, {'A': 1.075, 'B': 0.45}, {'A': 1.1, 'B': 0.447},
            {'A': 1.15, 'B': 0.437}, {'A': 1.2, 'B': 0.429}, {'A': 1.3, 'B': 0.418},
            {'A': 1.4, 'B': 0.406}, {'A': 1.5, 'B': 0.394}, {'A': 1.6, 'B': 0.382},
            {'A': 1.8, 'B': 0.359}, {'A': 2.0, 'B': 0.339}, {'A': 2.2, 'B': 0.321},
            {'A': 2.4, 'B': 0.301}, {'A': 2.6, 'B': 0.28}, {'A': 3.0, 'B': 0.25},
            {'A': 4.0, 'B': 0.2}, {'A': 5.0, 'B': 0.18}
        ]

        p = cProfile(drag_table=0, custom_drag_function=custom_drag_func)
        data = p.calculate_trajectory()

    def test_time(self):
        with self.subTest('def init') as st:
            print(timeit.timeit(lambda: cProfile(), number=1), 'def init')

        with self.subTest('def init + make'):
            p = cProfile()
            print(timeit.timeit(lambda: p.calculate_trajectory(), number=1), 'def init + make', )

        with self.subTest('max=2500m, step=1m'):
            p = cProfile(
                maximum_distance=(2500, unit.DistanceMeter),
                distance_step=(1, unit.DistanceMeter),
            )
            print(timeit.timeit(lambda: p.calculate_trajectory(), number=1), 'max=2500m, step=1m')

        with self.subTest('max=2500m, step=1m, max_step=5ft'):
            p = cProfile(
                maximum_distance=(2500, unit.DistanceMeter),
                distance_step=(1, unit.DistanceMeter),
                maximum_step_size=(5, unit.DistanceFoot)
            )
            print(timeit.timeit(lambda: p.calculate_trajectory(), number=1), 'max=2500m, step=1m, max_step=5ft')

        with self.subTest('custom_df'):
            print(timeit.timeit(self.test_custom_df, number=1), 'max=2500m, step=1m, max_step=5ft, custom_df')


class TestAtmo(unittest.TestCase):

    def test_create(self):
        v = Atmosphere(
            altitude=Distance(0, DistanceMeter),
            pressure=Pressure(760, PressureMmHg),
            temperature=Temperature(15, TemperatureCelsius),
            humidity=0.5
        )

        icao = IcaoAtmosphere(Distance(0, DistanceMeter))

    # @unittest.SkipTest
    def test_time(self):
        t = timeit.timeit(self.test_create, number=1)
        print(t)


class TestShotParams(unittest.TestCase):

    def test_create(self):
        v = ShotParameters(
            Angular(0, AngularDegree),
            Distance(1000, DistanceFoot),
            Distance(100, DistanceFoot)
        )

    def test_unlevel(self):
        v = ShotParametersUnlevel(
            Angular(0, AngularDegree),
            Distance(1000, DistanceFoot),
            Distance(100, DistanceFoot),
            Angular(0, AngularDegree),
            Angular(0, AngularDegree)
        )

    # @unittest.SkipTest
    def test_time(self):
        t = timeit.timeit(self.test_create, number=1)
        print(datetime.fromtimestamp(t).time().strftime('%S.%fs'))
        t = timeit.timeit(self.test_unlevel, number=1)
        print(datetime.fromtimestamp(t).time().strftime('%S.%fs'))


class TestDrag(unittest.TestCase):

    def setUp(self) -> None:
        self.bc = self.test_create()

    def test_create(self):
        bc = BallisticCoefficient(
            value=0.275,
            drag_table=DragTableG7,
            weight=Weight(178, WeightGrain),
            diameter=Distance(0.308, DistanceInch),
            custom_drag_table=[]
        )
        return bc

    def test_drag(self):
        return self.bc.drag(3)

    def test_time(self):
        t = timeit.timeit(self.test_create, number=1)
        print(datetime.fromtimestamp(t).time().strftime('%S.%fs'))
        t = timeit.timeit(self.test_drag, number=50000)
        print(datetime.fromtimestamp(t).time().strftime('%S.%fs'))


class TestG7Profile(unittest.TestCase):

    def test_drag(self):
        bc = BallisticCoefficient(
            value=0.223,
            drag_table=DragTableG7,
            weight=Weight(167, WeightGrain),
            diameter=Distance(0.308, DistanceInch),
            custom_drag_table=[]
        )

        print(bc.form_factor())
        print(bc.drag(3))

        ret = bc.calculated_drag_function()
        print(ret)

    def test_mbc(self):
        bc = MultipleBallisticCoefficient(
            drag_table=DragTableG7,
            weight=Weight(178, WeightGrain),
            diameter=Distance(0.308, DistanceInch),
            multiple_bc_table=[[0.275, 800], [0.255, 500], [0.26, 700], ],
            velocity_units_flag=VelocityMPS
        )

        ret = bc.custom_drag_func()
        print(ret)

    def test_create(self):
        bc = BallisticCoefficient(
            value=0.223,
            drag_table=DragTableG7,
            weight=Weight(167, WeightGrain),
            diameter=Distance(0.308, DistanceInch),
            custom_drag_table=[]
        )

        p1 = ProjectileWithDimensions(
            bc,
            Distance(0.308, DistanceInch),
            Distance(1.2, DistanceInch),
            Weight(167, WeightGrain),
        )

        ammo = Ammunition(p1, Velocity(800, VelocityMPS))
        atmo = Atmosphere(Distance(0, DistanceMeter), Pressure(760, PressureMmHg),
                          Temperature(15, TemperatureCelsius), 0.5)

        zero = ZeroInfo(Distance(100, DistanceMeter), True, True, ammo, atmo)
        twist = TwistInfo(TwistRight, Distance(11, DistanceInch))
        weapon = WeaponWithTwist(Distance(90, DistanceMillimeter), zero, twist)
        wind = create_only_wind_info(Velocity(0, VelocityMPS), Angular(0, AngularDegree))
        calc = TrajectoryCalculator()
        calc.set_maximum_calculator_step_size(Distance(1, DistanceFoot))
        print(timeit.timeit(lambda: calc.sight_angle(ammo, weapon, atmo), number=1))
        sight_angle = calc.sight_angle(ammo, weapon, atmo)
        shot_info = ShotParameters(sight_angle, Distance(2500, DistanceMeter), Distance(1, DistanceMeter))
        return calc.trajectory(ammo, weapon, atmo, shot_info, wind)

    def test_time(self):
        t = timeit.timeit(self.test_create, number=1)
        print(datetime.fromtimestamp(t).time().strftime('%S.%fs'))


class TestPyBallisticCalc(unittest.TestCase):

    def test_zero1(self):
        bc = BallisticCoefficient(0.365, DragTableG1)
        projectile = Projectile(bc, unit.Weight(69, unit.WeightGrain))
        ammo = Ammunition(projectile, unit.Velocity(2600, unit.VelocityFPS))
        zero = ZeroInfo(unit.Distance(100, unit.DistanceYard))
        weapon = Weapon(unit.Distance(3.2, unit.DistanceInch), zero)
        atmosphere = IcaoAtmosphere(Distance(0, DistanceMeter))
        calc = TrajectoryCalculator()

        sight_angle = calc.sight_angle(ammo, weapon, atmosphere)

        self.assertLess(fabs(sight_angle.get_in(unit.AngularRadian) - 0.001651), 1e-6,
                        f'TestZero1 failed {sight_angle.get_in(unit.AngularRadian):.10f}')

    def test_zero2(self):
        bc = BallisticCoefficient(0.223, DragTableG7)
        projectile = Projectile(bc, unit.Weight(168, unit.WeightGrain))
        ammo = Ammunition(projectile, unit.Velocity(2750, unit.VelocityFPS))
        zero = ZeroInfo(unit.Distance(100, unit.DistanceYard))
        weapon = Weapon(unit.Distance(2, unit.DistanceInch), zero)
        atmosphere = IcaoAtmosphere(Distance(0, DistanceMeter))
        calc = TrajectoryCalculator()

        sight_angle = calc.sight_angle(ammo, weapon, atmosphere)

        self.assertLess(fabs(sight_angle.get_in(unit.AngularRadian) - 0.001228), 1e-6,
                        f'TestZero2 failed {sight_angle.get_in(unit.AngularRadian):.10f}')

    def assertEqualCustom(self, a, b, accuracy, name):
        with self.subTest():
            self.assertFalse(fabs(a - b) > accuracy, f'Assertion {name} failed ({a}/{b}, {accuracy})')

    def validate_one(self, data: TrajectoryData, distance: float, velocity: float, mach: float, energy: float,
                     path: float, hold: float, windage: float, wind_adjustment: float, time: float, ogv: float,
                     adjustment_unit: int):

        self.assertEqualCustom(distance, data.travelled_distance().get_in(unit.DistanceYard), 0.001, "Distance")
        self.assertEqualCustom(velocity, data.velocity().get_in(unit.VelocityFPS), 5, "Velocity")
        self.assertEqualCustom(mach, data.mach_velocity(), 0.005, "Mach")
        self.assertEqualCustom(energy, data.energy().get_in(unit.EnergyFootPound), 5, "Energy")
        self.assertEqualCustom(time, data.time().total_seconds(), 0.06, "Time")
        self.assertEqualCustom(ogv, data.optimal_game_weight().get_in(unit.WeightPound), 1, "OGV")

        if distance >= 800:
            self.assertEqualCustom(path, data.drop().get_in(unit.DistanceInch), 4, 'Drop')
        elif distance >= 500:
            self.assertEqualCustom(path, data.drop().get_in(unit.DistanceInch), 1, 'Drop')
        else:
            self.assertEqualCustom(path, data.drop().get_in(unit.DistanceInch), 0.5, 'Drop')

        if distance > 1:
            self.assertEqualCustom(hold, data.drop_adjustment().get_in(adjustment_unit), 0.5, 'Hold')

        if distance >= 800:
            self.assertEqualCustom(windage, data.windage().get_in(unit.DistanceInch), 1.5, "Windage")
        elif distance >= 500:
            self.assertEqualCustom(windage, data.windage().get_in(unit.DistanceInch), 1, "Windage")
        else:
            self.assertEqualCustom(windage, data.windage().get_in(unit.DistanceInch), 0.5, "Windage")

        if distance > 1:
            self.assertEqualCustom(wind_adjustment, data.windage_adjustment().get_in(adjustment_unit), 0.5, "WAdj")

    def test_path_g1(self):
        bc = BallisticCoefficient(0.223, DragTableG1)
        projectile = Projectile(bc, unit.Weight(168, unit.WeightGrain))
        ammo = Ammunition(projectile, unit.Velocity(2750, unit.VelocityFPS))
        zero = ZeroInfo(unit.Distance(100, unit.DistanceYard))
        weapon = Weapon(unit.Distance(2, unit.DistanceInch), zero)
        atmosphere = IcaoAtmosphere(Distance(0, DistanceMeter))
        shot_info = ShotParameters(unit.Angular(0.001228, unit.AngularRadian),
                                   unit.Distance(1000, unit.DistanceYard),
                                   unit.Distance(100, unit.DistanceYard))
        wind = create_only_wind_info(unit.Velocity(5, unit.VelocityMPH),
                                     unit.Angular(-45, unit.AngularDegree))
        calc = TrajectoryCalculator()
        data = calc.trajectory(ammo, weapon, atmosphere, shot_info, wind)

        self.assertEqualCustom(len(data), 11, 0.1, "Length")

        test_data = [
            [data[0], 0, 2750, 2.463, 2820.6, -2, 0, 0, 0, 0, 880, unit.AngularMOA],
            [data[1], 100, 2351.2, 2.106, 2061, 0, 0, -0.6, -0.6, 0.118, 550, unit.AngularMOA],
            [data[5], 500, 1169.1, 1.047, 509.8, -87.9, -16.8, -19.5, -3.7, 0.857, 67, unit.AngularMOA],
            [data[10], 1000, 776.4, 0.695, 224.9, -823.9, -78.7, -87.5, -8.4, 2.495, 20, unit.AngularMOA]
        ]

        for d in test_data:
            with self.subTest():
                self.validate_one(*d)

    def test_path_g7(self):
        bc = BallisticCoefficient(0.223, DragTableG7,
                                  weight=Weight(167, WeightGrain),
                                  diameter=Distance(0.308, DistanceInch),
                                  custom_drag_table=[])
        projectile = ProjectileWithDimensions(bc, unit.Distance(0.308, unit.DistanceInch),
                                              unit.Distance(1.282, unit.DistanceInch),
                                              unit.Weight(168, unit.WeightGrain))
        ammo = Ammunition(projectile, unit.Velocity(2750, unit.VelocityFPS))
        zero = ZeroInfo(unit.Distance(100, unit.DistanceYard))
        twist = TwistInfo(TwistRight, unit.Distance(11.24, unit.DistanceInch))
        weapon = WeaponWithTwist(unit.Distance(2, unit.DistanceInch), zero, twist)
        atmosphere = IcaoAtmosphere(Distance(0, DistanceMeter))
        shot_info = ShotParameters(unit.Angular(4.221, unit.AngularMOA),
                                   unit.Distance(1000, unit.DistanceYard),
                                   unit.Distance(100, unit.DistanceYard))
        wind = create_only_wind_info(unit.Velocity(5, unit.VelocityMPH),
                                     unit.Angular(-45, unit.AngularDegree))

        calc = TrajectoryCalculator()
        data = calc.trajectory(ammo, weapon, atmosphere, shot_info, wind)

        self.assertEqualCustom(len(data), 11, 0.1, "Length")

        test_data = [
            [data[0], 0, 2750, 2.463, 2820.6, -2, 0, 0, 0, 0, 880, unit.AngularMil],
            [data[1], 100, 2544.3, 2.279, 2416, 0, 0, -0.35, -0.09, 0.113, 698, unit.AngularMil],
            [data[5], 500, 1810.7, 1.622, 1226, -56.3, -3.18, -9.96, -0.55, 0.673, 252, unit.AngularMil],
            [data[10], 1000, 1081.3, 0.968, 442, -401.6, -11.32, -50.98, -1.44, 1.748, 55, unit.AngularMil]
        ]

        for d in test_data:
            with self.subTest():
                self.validate_one(*d)



from libc.math cimport pow

from .drag import load_drag_table
from .bmath.unit import *
from .atmosphere import IcaoAtmosphere

cdef class BCDataPoint:
    cdef double _bc
    cdef double _v

    def __init__(self, bc: double, v: double):
        self._bc = bc
        self._v = v

    cpdef double bc(self):
        return self._bc

    cpdef double v(self):
        return self._v

    cpdef set_v(self, float value):
        self._v = value


cdef class MultipleBallisticCoefficient:
    cdef list _custom_drag_table, _table_data, _bc_table, _multiple_bc_table
    cdef double _sectional_density, _speed_of_sound
    cdef int _units, _table
    cdef _weight, _diameter

    def __init__(self, drag_table: int, diameter: Distance, weight: Weight,
                 multiple_bc_table: list[(double, double)], velocity_units_flag: int):

        cdef double altitude, density, mach
        cdef _atmosphere

        self._multiple_bc_table = multiple_bc_table
        self._table = drag_table
        self._weight = weight
        self._diameter = diameter
        self._units = velocity_units_flag
        self._sectional_density = self._get_sectional_density()

        _atmosphere = IcaoAtmosphere(Distance(0, DistanceFoot))

        altitude = Distance(0, DistanceMeter).get_in(DistanceFoot)
        density, mach = _atmosphere.get_density_factor_and_mach_for_altitude(altitude)
        self._speed_of_sound = Velocity(mach, VelocityFPS).get_in(VelocityMPS)

        self._table_data = load_drag_table(self._table)
        self._bc_table = self._create_bc_table_data_points()
        self._custom_drag_table = []

    cdef double _get_sectional_density(self):
        cdef double w, d
        w = self._weight.get_in(WeightGrain)
        d = self._diameter.get_in(DistanceInch)
        return w / pow(d, 2) / 7000

    cdef double _get_form_factor(self, double bc):
        return self._sectional_density / bc

    cdef list _bc_extended(self):
        cdef list bc_mah, bc_extended, df_part
        cdef bc_max, bc_min
        cdef int ddf
        cdef double bc_delta

        bc_mah = [BCDataPoint(point.bc(), point.v() / self._speed_of_sound) for point in self._bc_table]
        bc_mah.insert(len(bc_mah), BCDataPoint(bc_mah[-1].bc(), self._table_data[0].a()))
        bc_mah.insert(0, BCDataPoint(bc_mah[0].bc(), self._table_data[-1].a()))
        bc_extended = [bc_mah[0].bc(), ]

        for i in range(1, len(bc_mah)):
            bc_max = bc_mah[i - 1]
            bc_min = bc_mah[i]
            df_part = list(filter(lambda point: bc_max.v() > point.a() >= bc_min.v(), self._table_data))
            ddf = len(df_part)
            bc_delta = (bc_max.bc() - bc_min.bc()) / ddf
            for j in range(ddf):
                bc_extended.append(bc_max.bc() - bc_delta * j)

        return bc_extended

    cdef double _get_counted_cd(self, double form_factor, double cdst):
        return cdst * form_factor

    cdef list _create_bc_table_data_points(self):
        cdef list bc_table
        cdef data_point
        cdef double bc, v,
        self._multiple_bc_table.sort(reverse=True, key=lambda x: x[1])
        bc_table = []
        for bc, v in self._multiple_bc_table:
            data_point = BCDataPoint(bc, Velocity(v, self._units).get_in(VelocityMPS))
            bc_table.append(data_point)
        return bc_table

    cdef list _calculate_custom_drag_func(self):
        cdef list bc_extended, drag_function
        cdef int i
        cdef point
        cdef double form_factor, cd, bc

        bc_extended = self._bc_extended()
        drag_function = []
        for i, point in enumerate(self._table_data):
            bc = bc_extended[len(bc_extended) - 1 - i]
            form_factor = self._get_form_factor(bc)
            cd = form_factor * point.b()
            drag_function.append({'A': point.a(), 'B': cd})
        self._custom_drag_table = drag_function

    cpdef list custom_drag_func(self):
        self._calculate_custom_drag_func()
        return self._custom_drag_table



from libc.math cimport floor, pow
from .bmath.unit import *
from .drag_tables import *

DragTableG1: int = 1
DragTableG2: int = 2
DragTableG5: int = 3
DragTableG6: int = 4
DragTableG7: int = 5
DragTableG8: int = 6
DragTableGS: int = 7
DragTableGI: int = 8

cdef class BallisticCoefficient:
    cdef double _value
    cdef int _table
    cdef list _table_data
    cdef list _curve_data
    cdef _weight, _diameter
    cdef double _sectional_density, _form_factor
    cdef list _custom_drag_table

    def __init__(self, value: double, drag_table: int, weight: Weight, diameter: Distance, custom_drag_table: list):

        self._table = drag_table

        self._weight = weight
        self._diameter = diameter
        self._sectional_density = self._get_sectional_density()
        self._custom_drag_table = custom_drag_table

        if self._table == 0 and len(self._custom_drag_table) > 0:
            self._form_factor = 0.999  # defined as form factor in lapua-like custom CD data
            self._value = self._get_custom_bc()
            self._table_data = make_data_points(self._custom_drag_table)
            self._curve_data = calculate_curve(self._table_data)

        elif drag_table < DragTableG1 or DragTableG1 > DragTableGI:
            raise ValueError(f"BallisticCoefficient: Unknown drag table {drag_table}")
        elif value <= 0:
            raise ValueError('BallisticCoefficient: Drag coefficient must be greater than zero')
        elif self._table == 0 and len(custom_drag_table) == 0:
            raise ValueError('BallisticCoefficient: Custom drag table must be longer than 0')
        else:
            self._value = value
            self._form_factor = self._get_form_factor()
            self._table_data = load_drag_table(self._table)
            self._curve_data = calculate_curve(self._table_data)

    cpdef double drag(self, double mach):
        cdef double cd
        cd = calculate_by_curve(self._table_data, self._curve_data, mach)
        return cd * 2.08551e-04 / self._value

    cpdef double value(self):
        return self._value

    cpdef int table(self):
        return self._table

    cdef double _get_custom_bc(self):
        return self._sectional_density / self._form_factor

    cdef double _get_form_factor(self):
        return self._sectional_density / self._value

    cdef double _get_sectional_density(self):
        cdef double w, d
        w = self._weight.get_in(WeightGrain)
        d = self._diameter.get_in(DistanceInch)
        return w / pow(d, 2) / 7000

    cpdef double standard_cd(self, double mach):
        return calculate_by_curve(self._table_data, self._curve_data, mach)

    cpdef double calculated_cd(self, double mach):
        return self.standard_cd(mach) * self._form_factor

    cpdef list calculated_drag_function(self):
        cdef standard_cd_table
        cdef list calculated_cd_table
        cdef double st_mach, st_cd, cd

        calculated_cd_table = []

        for point in self._table_data:
            st_mach = point.a()
            st_cd = point.b()
            cd = self.calculated_cd(st_mach)
            calculated_cd_table.append({'A': st_mach, 'B': cd})

        return calculated_cd_table

    cpdef form_factor(self):
        return self._form_factor

cdef class DataPoint:
    cdef double _a, _b

    def __init__(self, a: double, b: double):
        self._a = a
        self._b = b

    cpdef double a(self):
        return self._a

    cpdef double b(self):
        return self._b

cdef class CurvePoint:
    cdef double _a, _b, _c

    def __init__(self, a: double, b: double, c: double):
        self._a = a
        self._b = b
        self._c = c

    cpdef double a(self):
        return self._a

    cpdef double b(self):
        return self._b

    cpdef double c(self):
        return self._c

cpdef list make_data_points(drag_table: list):
    table: list = []
    cdef data_point
    for point in drag_table:
        data_point = DataPoint(point['A'], point['B'])
        table.append(data_point)
    return table

cpdef list calculate_curve(list data_points):
    cdef double rate, x1, x2, x3, y1, y2, y3, a, b, c
    cdef curve = []
    cdef curve_point
    cdef int num_points, len_data_points, len_data_range

    rate = (data_points[1].b() - data_points[0].b()) / (data_points[1].a() - data_points[0].a())
    curve = [CurvePoint(0, rate, data_points[0].b() - data_points[0].a() * rate)]
    len_data_points = int(len(data_points))
    len_data_range = len_data_points - 1

    for i in range(1, len_data_range):
        x1 = data_points[i - 1].a()
        x2 = data_points[i].a()
        x3 = data_points[i + 1].a()
        y1 = data_points[i - 1].b()
        y2 = data_points[i].b()
        y3 = data_points[i + 1].b()
        a = ((y3 - y1) * (x2 - x1) - (y2 - y1) * (x3 - x1)) / (
                    (x3 * x3 - x1 * x1) * (x2 - x1) - (x2 * x2 - x1 * x1) * (x3 - x1))
        b = (y2 - y1 - a * (x2 * x2 - x1 * x1)) / (x2 - x1)
        c = y1 - (a * x1 * x1 + b * x1)
        curve_point = CurvePoint(a, b, c)
        curve.append(curve_point)

    num_points = len_data_points
    rate = (data_points[num_points - 1].b() - data_points[num_points - 2].b()) / \
           (data_points[num_points - 1].a() - data_points[num_points - 2].a())
    curve_point = CurvePoint(0, rate, data_points[num_points - 1].b() - data_points[num_points - 2].a() * rate)
    curve.append(curve_point)
    return curve

cpdef list load_drag_table(drag_table: int):
    cdef table

    if drag_table == DragTableG1:
        table = make_data_points(TableG1)
    elif drag_table == DragTableG2:
        table = make_data_points(TableG2)
    elif drag_table == DragTableG5:
        table = make_data_points(TableG5)
    elif drag_table == DragTableG6:
        table = make_data_points(TableG6)
    elif drag_table == DragTableG7:
        table = make_data_points(TableG7)
    elif drag_table == DragTableG8:
        table = make_data_points(TableG8)
    elif drag_table == DragTableGI:
        table = make_data_points(TableGI)
    elif drag_table == DragTableGS:
        table = make_data_points(TableGS)
    else:
        raise ValueError("Unknown drag table type")
    return table

cpdef double calculate_by_curve(data: list, curve: list, mach: double):
    cdef int num_points, mlo, mhi, mid

    num_points = int(len(curve))
    mlo = 0
    mhi = num_points - 2

    while mhi - mlo > 1:
        mid = int(floor(mhi + mlo) / 2.0)
        if data[mid].a() < mach:
            mlo = mid
        else:
            mhi = mid

    if data[mhi].a() - mach > mach - data[mlo].a():
        m = mlo
    else:
        m = mhi
    return curve[m].c() + mach * (curve[m].b() + curve[m].a() * mach)



from libc.math cimport pow, sqrt, fabs
from .bmath.unit import *


cIcaoStandardTemperatureR: double = 518.67
cIcaoFreezingPointTemperatureR: double = 459.67
cTemperatureGradient: double = -3.56616e-03
cIcaoStandardHumidity: double = 0.0
cPressureExponent: double = -5.255876
cSpeedOfSound: double = 49.0223
cA0: double = 1.24871
cA1: double = 0.0988438
cA2: double = 0.00152907
cA3: double = -3.07031e-06
cA4: double = 4.21329e-07
cA5: double = 3.342e-04
cStandardTemperature: double = 59.0
cStandardPressure: double = 29.92
cStandardDensity: double = 0.076474


cdef class Atmosphere:
    cdef double _density, _humidity, _mach1
    cdef _mach
    cdef _altitude
    cdef _pressure
    cdef _temperature

    def __init__(self, altitude: Distance, pressure: Pressure, temperature: Temperature, humidity: double):

        if humidity > 1:
            humidity = humidity / 100

        if humidity < 0 or humidity > 100:
            self.create_default()
        elif not altitude or not pressure or not temperature:
            self.create_default()
        else:
            self._altitude = altitude
            self._pressure = pressure
            self._temperature = temperature
            self._humidity = humidity

        self.calculate()

    def __str__(self) -> str:
        return f'Altitude: {self._altitude}, Pressure: {self._pressure}, ' \
               f'Temperature: {self._temperature}, Humidity: {self.humidity_in_percent:.2f}'

    cdef string(self):
        return self.__str__()

    cpdef create_default(self):
        self._altitude = Distance(0.0, DistanceFoot)
        self._pressure = Pressure(cStandardPressure, PressureInHg)
        self._temperature = Temperature(cStandardTemperature, TemperatureFahrenheit)
        self._humidity = 0.78

    cpdef altitude(self):
        return self._altitude

    cpdef temperature(self):
        return self._temperature

    cpdef pressure(self):
        return self._pressure

    cpdef double humidity(self):
        return self._humidity

    cpdef double humidity_in_percent(self):
        return self._humidity * 100

    cpdef double density(self):
        return self._density

    cpdef double density_factor(self):
        return self._density / cStandardDensity

    cpdef mach(self):
        return self._mach

    cdef (double, double) calculate0(self, double t, double p):
        cdef double et0, et, hc, density, mach

        if t > 0:
            et0 = cA0 + t * (cA1 + t * (cA2 + t * (cA3 + t * cA4)))
            et = cA5 * self._humidity * et0
            hc = (p - 0.3783 * et) / cStandardPressure
        else:
            hc = 1.0

        density = cStandardDensity * (cIcaoStandardTemperatureR / (t + cIcaoFreezingPointTemperatureR)) * hc
        mach = sqrt(t + cIcaoFreezingPointTemperatureR) * cSpeedOfSound
        return density, mach

    cdef calculate(self):
        cdef double density, mach, mach1, t, p
        t = self._temperature.get_in(TemperatureFahrenheit)
        p = self._pressure.get_in(PressureInHg)
        density, mach = self.calculate0(t, p)
        self._density = density
        self._mach1 = mach
        self._mach = Velocity(mach, VelocityFPS)

    cpdef (double, double) get_density_factor_and_mach_for_altitude(self, altitude: double):
        cdef double density, mach, t0, p, ta, tb, t, org_altitude
        org_altitude = self._altitude.get_in(DistanceFoot)
        if fabs(org_altitude - altitude) < 30:
            density = self._density / cStandardDensity
            mach = self._mach1
            return density, mach

        t0 = self._temperature.get_in(TemperatureFahrenheit)
        p = self._pressure.get_in(PressureInHg)

        ta = cIcaoStandardTemperatureR + org_altitude * cTemperatureGradient - cIcaoFreezingPointTemperatureR
        tb = cIcaoStandardTemperatureR + altitude * cTemperatureGradient - cIcaoFreezingPointTemperatureR
        t = t0 + ta - tb
        p = p * pow(t0 / t, cPressureExponent)

        density, mach = self.calculate0(t, p)
        return density / cStandardDensity, mach


cpdef IcaoAtmosphere(altitude: Distance):
    cdef temperature, pressure
    temperature = Temperature(
        cIcaoStandardTemperatureR + altitude.get_in(DistanceFoot)
        * cTemperatureGradient - cIcaoFreezingPointTemperatureR, TemperatureFahrenheit)

    pressure = Pressure(
        cStandardPressure *
        pow(cIcaoStandardTemperatureR / (
                temperature.get_in(TemperatureFahrenheit) + cIcaoFreezingPointTemperatureR),
                 cPressureExponent),
        PressureInHg)

    return Atmosphere(altitude, pressure, temperature, cIcaoStandardHumidity)



from libc.math cimport sqrt, fabs

cdef struct vector:
    double x
    double y
    double z


cdef class Vector:
    cdef double _x
    cdef double _y
    cdef double _z

    def __init__(self, x: double, y: double, z: double):
        self._x = x
        self._y = y
        self._z = z

    def __str__(self):
        return f'{vector(self._x, self._y, self._z)}'

    cpdef double x(self):
        return self._x

    cpdef double y(self):
        return self._y

    cpdef double z(self):
        return self._z

    cdef string(self):
        cdef v = vector(self._x, self._y, self._z)
        return f'{v}'

    cpdef Vector copy(self):
        return Vector(self._x, self._y, self._z)

    cpdef double magnitude(self):
        cdef double m = sqrt(self._x * self._x + self._y * self._y + self._z * self._z)
        return m

    cpdef Vector multiply_by_const(self, float a):
        return Vector(self._x * a, self._y * a, self._z * a)

    cpdef double multiply_by_vector(self, b: Vector):
        cdef double var = self._x * b._x + self._y * b._y + self._z * b._z
        return var

    cpdef Vector add(self, b: Vector):
        return Vector(self._x + b._x, self._y + b._y, self._z + b._z)

    cpdef Vector subtract(self, b: Vector):
        return Vector(self._x - b._x, self._y - b._y, self._z - b._z)

    cpdef Vector negate(self):
        return Vector(-self._x, -self._y, -self._z)

    cpdef Vector normalize(self):
        cdef double m = self.magnitude()
        if fabs(m) < 1e-10:
            return Vector(self._x, self._y, self._z)
        return self.multiply_by_const(1.0 / m)

# cython
# test_time_1 0.5331406999999999
# test_time_2 1.0104932999999998
# test_time_3 0.4101440999999997

# pure python
# test_time_1 0.6663539999999999
# test_time_2 1.3299572
# test_time_3 0.48653330000000006



from .atmosphere import Atmosphere
from .drag import BallisticCoefficient, DragTableG7
from .projectile import ProjectileWithDimensions
from .weapon import Ammunition, ZeroInfo, TwistInfo, TwistRight, WeaponWithTwist
from .wind import create_only_wind_info
from .shot_parameters import ShotParametersUnlevel
from .trajectory_calculator import TrajectoryCalculator
from .bmath.unit import *
from .multiple_bc import MultipleBallisticCoefficient

cdef class Profile(object):
    cdef int _drag_table, _twist_direction
    cdef list _custom_drag_function
    cdef list _calculated_drag_function
    cdef list _trajectory_data
    cdef list _multiple_bc_table
    cdef double _humidity
    cdef _bc_value, _bullet_diameter, _bullet_length, _bullet_weight
    cdef _muzzle_velocity, _zero_distance, _maximum_distance, _distance_step
    cdef _altitude, _pressure, _temperature
    cdef _twist, _sight_height, _sight_angle
    cdef _wind_velocity, _wind_direction
    cdef _maximum_step_size, _shot_angle, _cant_angle

    def __init__(self,
                 bc_value: double = 0.223,
                 drag_table: int = DragTableG7,
                 bullet_diameter: (double, int) = (0.308, DistanceInch),
                 bullet_length: (double, int) = (1.2, DistanceInch),
                 bullet_weight: (double, int) = (167, WeightGrain),
                 muzzle_velocity: (double, int) = (800, VelocityMPS),
                 altitude: (double, int) = (0, DistanceMeter),
                 pressure: (double, int) = (760, PressureMmHg),
                 temperature: (double, int) = (15, TemperatureCelsius),
                 humidity: double = 0.5,
                 zero_distance: (double, int) = (100, DistanceMeter),
                 twist: (double, int) = (11, DistanceInch),
                 twist_direction: int = TwistRight,
                 sight_height: (double, int) = (90, DistanceMillimeter),
                 sight_angle: (double, int) = (0, AngularMOA),
                 maximum_distance: (double, int) = (1001, DistanceMeter),
                 distance_step: (double, int) = (100, DistanceMeter),
                 wind_velocity: (double, int) = (0, VelocityKMH),
                 wind_direction: (double, int) = (0, AngularDegree),
                 maximum_step_size: (double, int) = (1, DistanceFoot),
                 shot_angle: (double, int) = (0, AngularRadian),
                 cant_angle: (double, int) = (0, AngularRadian),
                 custom_drag_function=None,
                 multiple_bc_table=None
                 ):
        if custom_drag_function is None:
            custom_drag_function = []
        if multiple_bc_table is None:
            multiple_bc_table = []
        self._bc_value = bc_value
        self._drag_table = drag_table
        self._bullet_diameter = Distance(*bullet_diameter)
        self._bullet_length = Distance(*bullet_length)
        self._bullet_weight = Weight(*bullet_weight)
        self._muzzle_velocity = Velocity(*muzzle_velocity)
        self._altitude = Distance(*altitude)
        self._pressure = Pressure(*pressure)
        self._temperature = Temperature(*temperature)
        self._humidity = humidity
        self._zero_distance = Distance(*zero_distance)
        self._twist = Distance(*twist)
        self._twist_direction = twist_direction
        self._sight_height = Distance(*sight_height)
        self._sight_angle = Angular(*sight_angle)
        self._maximum_distance = Distance(*maximum_distance)
        self._distance_step = Distance(*distance_step)
        self._wind_velocity = Velocity(*wind_velocity)
        self._wind_direction = Angular(*wind_direction)
        self._maximum_step_size = Distance(*maximum_step_size)
        self._shot_angle = Angular(*shot_angle)
        self._cant_angle = Angular(*cant_angle)
        self._multiple_bc_table = multiple_bc_table
        self._trajectory_data = []
        self._custom_drag_function = custom_drag_function
        self._calculated_drag_function = []

    def dict(self):
        profile = {
            'drag_table': self._drag_table,
            'twist_direction': self._twist_direction,
            'custom_drag_function': self._custom_drag_function,
            'calculated_drag_function': self._calculated_drag_function,
            'humidity': self._humidity,
            'bc_value': self._bc_value,
            'bullet_diameter': self._bullet_diameter,
            'bullet_length': self._bullet_length,
            'bullet_weight': self._bullet_weight,
            'muzzle_velocity': self._muzzle_velocity,
            'zero_distance': self._zero_distance,
            'maximum_distance': self._maximum_distance,
            'distance_step': self._distance_step,
            'altitude': self._altitude,
            'pressure': self._pressure,
            'temperature': self._temperature,
            'twist': self._twist,
            'sight_height': self._sight_height,
            'sight_angle': self._sight_angle,
            'wind_velocity': self._wind_velocity,
            'wind_direction': self._wind_direction,
            'maximum_step_size': self._maximum_step_size,
            'shot_angle': self._shot_angle,
            'cant_angle': self._cant_angle,
            'multiple_bc_table': self._multiple_bc_table,
        }
        return profile

    def calculate_trajectory(self):
        self.make_calculator()
        return self._trajectory_data

    def calculate_drag_table(self):
        self.make_drag_table()
        return self._calculated_drag_function

    cdef make_bc(self):
        drag_table = self._drag_table

        if len(self._multiple_bc_table) > 0 >= self._bc_value:
            mbc = MultipleBallisticCoefficient(
                self._drag_table,
                self._bullet_diameter,
                self._bullet_weight,
                self._multiple_bc_table,
                self._muzzle_velocity.units()
            )
            self._custom_drag_function = mbc.custom_drag_func()
            self._bc_value = 0

            drag_table = 0

        return BallisticCoefficient(self._bc_value, drag_table,
                                    self._bullet_weight, self._bullet_diameter,
                                    self._custom_drag_function)

    cdef make_drag_table(self):
        cdef bc
        bc = self.make_bc()
        self._calculated_drag_function = bc.calculated_drag_function()

    cdef make_calculator(self):
        cdef bc, projectile, ammo, atmo, zero, twist, weapon, wind, calc, angle, shot, data
        bc = self.make_bc()
        projectile = ProjectileWithDimensions(bc, self._bullet_diameter, self._bullet_length, self._bullet_weight)
        ammo = Ammunition(projectile, self._muzzle_velocity)
        atmo = Atmosphere(self._altitude, self._pressure, self._temperature, self._humidity)
        zero = ZeroInfo(self._zero_distance, True, True, ammo, atmo)
        twist = TwistInfo(self._twist_direction, self._twist)
        weapon = WeaponWithTwist(self._sight_height, zero, twist)
        wind = create_only_wind_info(self._wind_velocity, self._wind_direction)
        calc = TrajectoryCalculator()
        calc.set_maximum_calculator_step_size(self._maximum_step_size)
        angle = calc.sight_angle(ammo, weapon, atmo)
        shot = ShotParametersUnlevel(angle, self._maximum_distance, self._distance_step,
                                     self._shot_angle, self._cant_angle)
        data = calc.trajectory(ammo, weapon, atmo, shot, wind)
        self._trajectory_data = data

    cpdef double bc_value(self):
        return self._bc_value

    cpdef set_bc_value(self, value: double):
        self._bc_value = value

    cpdef int drag_table(self):
        return self._drag_table

    cpdef set_drag_table(self, drag_table: int):
        self._drag_table = drag_table

    cpdef bullet_diameter(self):
        return self._bullet_diameter

    cpdef set_bullet_diameter(self, value: double, units: int):
        self._bullet_diameter = Distance(value, units)

    cpdef bullet_length(self):
        return self._bullet_length

    cpdef set_bullet_length(self, value: double, units: int):
        self._bullet_length = Distance(value, units)

    cpdef bullet_weight(self):
        return self._bullet_weight

    cpdef set_bullet_weight(self, value: double, units: int):
        self._bullet_weight = Weight(value, units)

    cpdef muzzle_velocity(self):
        return self._muzzle_velocity

    cpdef set_muzzle_velocity(self, value: double, units: int):
        self._muzzle_velocity = Velocity(value, units)

    cpdef altitude(self):
        return self._altitude

    cpdef set_altitude(self, value: double, units: int):
        self._altitude = Distance(value, units)

    cpdef pressure(self):
        return self._pressure

    cpdef set_pressure(self, value: double, units: int):
        self._pressure = Pressure(value, units)

    cpdef temperature(self):
        return self._temperature

    cpdef set_temperature(self, value: double, units: int):
        self._temperature = Temperature(value, units)

    cpdef double humidity(self):
        return self._humidity

    cpdef set_humidity(self, value: double):
        self._humidity = value

    cpdef zero_distance(self):
        return self._zero_distance

    cpdef set_zero_distance(self, value: double, units: int):
        self._zero_distance = Distance(value, units)

    cpdef twist(self):
        return self._twist

    cpdef set_twist(self, value: double, units: int):
        self._twist = Distance(value, units)

    cpdef int twist_direction(self):
        return self._twist_direction

    cpdef set_twist_direction(self, direction: int):
        self._twist_direction = direction

    cpdef sight_height(self):
        return self._sight_height

    cpdef set_sight_height(self, value: double, units: int):
        self._sight_height = Distance(value, units)

    cpdef sight_angle(self):
        return self._sight_angle

    cpdef set_sight_angle(self, value: double, units: int):
        self._sight_angle = Angular(value, units)

    cpdef maximum_distance(self):
        return self._maximum_distance

    cpdef set_maximum_distance(self, value: double, units: int):
        self._maximum_distance = Distance(value, units)

    cpdef distance_step(self):
        return self._distance_step

    cpdef set_distance_step(self, value: double, units: int):
        self._distance_step = Distance(value, units)

    cpdef wind_velocity(self):
        return self._wind_velocity

    cpdef set_wind_velocity(self, value: double, units: int):
        self._wind_velocity = Velocity(value, units)

    cpdef wind_direction(self):
        return self._wind_direction

    cpdef set_wind_direction(self, value: double, units: int):
        self._wind_direction = Angular(value, units)

    cpdef custom_drag_function(self):
        return self._custom_drag_function

    cpdef set_custom_drag_function(self, data: list[dict[str, double]]):
        self._custom_drag_function = data
        self._drag_table = 0
        self._bc_value = 0

    cpdef maximum_step_size(self):
        return self._maximum_step_size

    cpdef set_maximum_step_size(self, value: double, units: int):
        self._maximum_step_size = Distance(value, units)



TemperatureFahrenheit: int = 50
TemperatureCelsius: int = 51
TemperatureKelvin: int = 52
TemperatureRankin: int = 53


cdef class Temperature:
    cdef double _value
    cdef int _default_units
    cdef __name__

    def __init__(self, value: double, units: int):
        self.__name__ = 'Temperature'
        self._value = self.to_default(value, units)
        self._default_units = units

    cdef double to_default(self, value: double, units: int):
            if units == TemperatureFahrenheit:
                return value
            elif units == TemperatureRankin:
                return value - 459.67
            elif units == TemperatureCelsius:
                return value * 9 / 5 + 32
            elif units == TemperatureKelvin:
                return (value - 273.15) * 9 / 5 + 32
            else:
                raise KeyError(f'{self.__name__}: unit {units} is not supported')

    cdef double from_default(self, value: double, units: int):
        if units == TemperatureFahrenheit:
            return value
        elif units == TemperatureRankin:
            return value + 459.67
        elif units == TemperatureCelsius:
            return (value - 32) * 5 / 9
        elif units == TemperatureKelvin:
            return (value - 32) * 5 / 9 + 273.15
        else:
            raise KeyError(f'KeyError: {self.__name__}: unit {units} is not supported')

    cpdef double value(self, units: int):
        return self.from_default(self._value, units)

    cpdef Temperature convert(self, units: int):
        cdef double value = self.get_in(units)
        return Temperature(value, units)

    cpdef double get_in(self, units: int):
        return self.from_default(self._value, units)

    def __str__(self):
        return self.string()

    cdef string(self):
        cdef name
        cdef int accuracy
        cdef int default = self._default_units
        cdef double v = self.from_default(self._value, default)
        if default == TemperatureFahrenheit:
            name = '°F'
            accuracy = 1
        elif default == TemperatureRankin:
            name = '°R'
            accuracy = 1
        elif default == TemperatureCelsius:
            name = '°C'
            accuracy = 1
        elif default == TemperatureKelvin:
            name = '°K'
            accuracy = 1
        else:
            name = '?'
            accuracy = 6
        return f'{round(v, accuracy)} {name}'

    cpdef int units(self):
        return self._default_units



from libc.math cimport fmod, floor
from .bmath.unit import *


cdef class Timespan:
    cdef double _time

    def __init__(self, time: double):
        self._time = time

    cpdef double total_seconds(self):
        return self._time

    cpdef seconds(self):
        return fmod(floor(self._time), 60)

    cpdef minutes(self):
        return fmod(floor(self._time / 60), 60)


cdef class TrajectoryData:
    cdef _time
    cdef _travel_distance
    cdef _velocity
    cdef double _mach
    cdef _drop
    cdef _drop_adjustment
    cdef _windage
    cdef _windage_adjustment
    cdef _energy
    cdef _optimal_game_weight

    def __init__(self, time: Timespan,
                 travel_distance: Distance,
                 velocity: Velocity,
                 mach: double,
                 drop: Distance,
                 drop_adjustment: [Angular, None],
                 windage: Distance,
                 windage_adjustment: [Angular, None],
                 energy: Energy,
                 optimal_game_weight: Weight):
        self._time = time
        self._travel_distance = travel_distance
        self._velocity = velocity
        self._mach = mach
        self._drop = drop
        self._drop_adjustment = drop_adjustment
        self._windage = windage
        self._windage_adjustment = windage_adjustment
        self._energy = energy
        self._optimal_game_weight = optimal_game_weight

    cpdef time(self):
        return self._time

    cpdef travelled_distance(self):
        return self._travel_distance

    cpdef velocity(self):
        return self._velocity

    cpdef double mach_velocity(self):
        return self._mach

    cpdef drop(self):
        return self._drop

    cpdef drop_adjustment(self):
        return self._drop_adjustment

    cpdef windage(self):
        return self._windage

    cpdef windage_adjustment(self):
        return self._windage_adjustment

    cpdef energy(self):
        return self._energy

    def optimal_game_weight(self):
        return self._optimal_game_weight



#!/usr/bin/env python
import io
import os
import re
from pathlib import Path
from setuptools import setup, find_packages, Extension

try:
    from Cython.Build import cythonize
    USE_CYTHON = True
except ImportError:
    USE_CYTHON = False


def read(*names, **kwargs):
    try:
        with io.open(
                os.path.join(os.path.dirname(__file__), *names),
                encoding=kwargs.get("encoding", "utf8")
        ) as fp:
            return fp.read()
    except IOError:
        return ''


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def iter_extensions(path):
    founded_extensions = []
    extensions_dir = Path(path).parent
    for ext_path in Path.iterdir(extensions_dir):
        if ext_path.suffix == '.pyx':
            ext_name = f"{'.'.join(extensions_dir.parts)}.{ext_path.stem}"
            ext = Extension(ext_name, [ext_path.as_posix()])
            founded_extensions.append(ext)
    return founded_extensions


def no_cythonize(extensions, **_ignore):
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)

            if ext in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions


# extensions_paths = [
#     Extension('*', ['py_ballisticcalc/*.pyx']),
#     Extension('*', ['py_ballisticcalc/bmath/unit/*.pyx']),
#     Extension('*', ['py_ballisticcalc/bmath/vector/*.pyx']),
# ]

extensions_paths = [
    'py_ballisticcalc/*.pyx',
    'py_ballisticcalc/bmath/unit/*.pyx',
    'py_ballisticcalc/bmath/vector/*.pyx',
]

extensions = []
for path in extensions_paths:
    extensions += iter_extensions(path)


# CYTHONIZE = bool(int(os.getenv("CYTHONIZE", 0))) and use_cython is not None

# if CYTHONIZE:
if USE_CYTHON:
    compiler_directives = {"language_level": 3, "embedsignature": True}
    extensions = cythonize(extensions, compiler_directives=compiler_directives)
else:
    extensions = no_cythonize(extensions)

with open("requirements.txt") as fp:
    install_requires = fp.read().strip().split("\n")
    print(install_requires)

with open("requirements-dev.txt") as fp:
    dev_requires = fp.read().strip().split("\n")

setup(

    name='py_ballisticcalc',
    ext_modules=extensions,
    install_requires=install_requires,
    setup_requires=[
        'setuptools>=18.0',  # automatically handles Cython extensions
        'cython>=3.0.0a10',
    ],

    extras_require={
        "dev": dev_requires,
        "docs": ["sphinx", "sphinx-rtd-theme"]
    },

    version=find_version('py_ballisticcalc', '__init__.py'),
    url='https://github.com/o-murphy/py_ballisticcalc',
    download_url='https://pypi.python.org/pypi/py_ballisticcalc/',
    license='LGPL-3.0',
    author="o-murphy",
    author_email='thehelixpg@gmail.com',
    description='LGPL library for small arms ballistic calculations (Python 3)',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    zip_safe=True,
    py_modules=find_packages() + [
        'py_ballisticcalc.drag_tables',
        'py_ballisticcalc.bin_test',
        'py_ballisticcalc.bmath.unit.unit_test',
        'py_ballisticcalc.bmath.vector.vector_test',
    ],
)



from .bmath.unit import *


cdef class ShotParameters:
    cdef _sight_angle, _shot_angle, _cant_angle, _maximum_distance, _step

    def __init__(self, sight_angle: Angular, maximum_distance: Distance, step: Distance):
        self._sight_angle = sight_angle
        self._shot_angle = Angular(0, AngularRadian)
        self._cant_angle = Angular(0, AngularRadian)
        self._maximum_distance: Distance = maximum_distance
        self._step: Distance = step

    cpdef sight_angle(self):
        return self._sight_angle

    cpdef shot_angle(self):
        return self._shot_angle

    cpdef cant_angle(self):
        return self._cant_angle

    cpdef maximum_distance(self):
        return self._maximum_distance

    cpdef step(self):
        return self._step


cdef class ShotParametersUnlevel(ShotParameters):

    def __init__(self, sight_angle: Angular, maximum_distance: Distance,
                 step: Distance, shot_angle: Angular, cant_angle: Angular):
        super(ShotParametersUnlevel, self).__init__(sight_angle, maximum_distance, step)
        self._sight_angle = sight_angle
        self._shot_angle = shot_angle
        self._cant_angle = cant_angle
        self._maximum_distance: Distance = maximum_distance
        self._step: Distance = step



from .bmath.unit import *
from .atmosphere import Atmosphere
from .projectile import Ammunition

cdef class ZeroInfo:
    cdef _has_ammunition
    cdef _has_atmosphere
    cdef _zero_distance
    cdef _ammunition
    cdef _zero_atmosphere

    def __init__(self, distance: Distance,
                 has_ammunition: bool = False,
                 has_atmosphere: bool = False,
                 ammunition: Ammunition = None, atmosphere: Atmosphere = None):
        self._has_ammunition = has_ammunition
        self._has_atmosphere = has_atmosphere
        self._zero_distance = distance
        self._ammunition = ammunition
        self._zero_atmosphere = atmosphere

    cpdef has_ammunition(self):
        return self._has_ammunition

    cpdef ammunition(self):
        return self._ammunition

    cpdef has_atmosphere(self):
        return self._has_atmosphere

    cpdef atmosphere(self):
        return self._zero_atmosphere

    cpdef zero_distance(self):
        return self._zero_distance

cdef class ZeroInfoDef(ZeroInfo):
    def __init__(self, distance: Distance):
        super(ZeroInfoDef, self).__init__(
            has_ammunition=False,
            has_atmosphere=False,
            distance=distance
        )

cdef class ZeroInfoWithAtmosphere(ZeroInfo):
    def __init__(self, distance: Distance, atmosphere: Atmosphere):
        super(ZeroInfoWithAtmosphere, self).__init__(
            has_ammunition=False,
            has_atmosphere=True,
            distance=distance,
            atmosphere=atmosphere
        )

cdef class ZeroInfoWithAmmo(ZeroInfo):
    def __init__(self, distance: Distance, ammo: Ammunition):
        super(ZeroInfoWithAmmo, self).__init__(
            has_ammunition=True,
            has_atmosphere=False,
            distance=distance,
            ammunition=ammo
        )

cdef class ZeroInfoWithAmmoAndAtmo(ZeroInfo):
    def __init__(self, distance: Distance, ammo: Ammunition, atmosphere: Atmosphere):
        super(ZeroInfoWithAmmoAndAtmo, self).__init__(
            has_ammunition=True,
            has_atmosphere=False,
            distance=distance,
            ammunition=ammo,
            atmosphere=atmosphere
        )

TwistRight: int = 1
TwistLeft: int = 2

cdef class TwistInfo:
    cdef int _twist_direction
    cdef _rifling_twist

    def __init__(self, direction: int, twist: Distance):
        self._twist_direction = direction
        self._rifling_twist = twist

    cpdef direction(self):
        return self._twist_direction

    cpdef twist(self):
        return self._rifling_twist

cdef class Weapon:
    cdef _sight_height
    cdef _zero_info
    cdef _has_twist_info
    cdef _twist
    cdef _click_value

    def __init__(self, sight_height: Distance, zero_info: ZeroInfo,
                 has_twist_info: bool = False, twist: TwistInfo = None, click_value: Angular = None):
        self._sight_height = sight_height
        self._zero_info = zero_info
        self._has_twist_info = has_twist_info
        self._twist = twist
        self._click_value = click_value

    cpdef sight_height(self):
        return self._sight_height

    cpdef zero(self):
        return self._zero_info

    cpdef has_twist(self):
        return self._has_twist_info

    cpdef twist(self):
        return self._twist

    cpdef click_value(self):
        return self._click_value

    cpdef set_click_value(self, click: Angular):
        self._click_value = click

cdef class WeaponWithTwist(Weapon):
    def __init__(self, sight_height: Distance, zero_info: ZeroInfo, twist: TwistInfo):
        super(WeaponWithTwist, self).__init__(sight_height, zero_info, True, twist)



DistanceInch = 10
DistanceFoot = 11
DistanceYard = 12
DistanceMile = 13
DistanceNauticalMile = 14
DistanceMillimeter = 15
DistanceCentimeter = 16
DistanceMeter = 17
DistanceKilometer = 18
DistanceLine = 19

cdef class Distance:
    cdef double _value
    cdef int _default_units
    cdef __name__

    def __init__(self, value: double, units: int):
        self.__name__ = 'Distance'
        self._value = self.to_default(value, units)
        self._default_units = units

    cdef double to_default(self, value: double, units: int):
        if units == DistanceInch:
            return value
        elif units == DistanceFoot:
            return value * 12
        elif units == DistanceYard:
            return value * 36
        elif units == DistanceMile:
            return value * 63360
        elif units == DistanceNauticalMile:
            return value * 72913.3858
        elif units == DistanceLine:
            return value / 10
        elif units == DistanceMillimeter:
            return value / 25.4
        elif units == DistanceCentimeter:
            return value / 2.54
        elif units == DistanceMeter:
            return value / 25.4 * 1000
        elif units == DistanceKilometer:
            return value / 25.4 * 1000000
        else:
            raise KeyError(f'{self.__name__}: unit {units} is not supported')

    cdef double from_default(self, value: double, units: int):
        if units == DistanceInch:
            return value
        elif units == DistanceFoot:
            return value / 12
        elif units == DistanceYard:
            return value / 36
        elif units == DistanceMile:
            return value / 63360
        elif units == DistanceNauticalMile:
            return value / 72913.3858
        elif units == DistanceLine:
            return value * 10
        elif units == DistanceMillimeter:
            return value * 25.4
        elif units == DistanceCentimeter:
            return value * 2.54
        elif units == DistanceMeter:
            return value * 25.4 / 1000
        elif units == DistanceKilometer:
            return value * 25.4 / 1000000
        else:
            raise KeyError(f'KeyError: {self.__name__}: unit {units} is not supported')

    cpdef double value(self, units: int):
        return self.from_default(self._value, units)

    cpdef Distance convert(self, units: int):
        cdef double value = self.get_in(units)
        return Distance(value, units)

    cpdef double get_in(self, units: int):
        return self.from_default(self._value, units)

    def __str__(self):
        return self.string()

    cdef string(self):
        cdef name
        cdef int accuracy
        cdef int default = self._default_units
        cdef double v = self.from_default(self._value, default)
        if default == DistanceInch:
            name = 'in'
            accuracy = 1
        elif default == DistanceFoot:
            name = 'ft'
            accuracy = 2
        elif default == DistanceYard:
            name = 'yd'
            accuracy = 3
        elif default == DistanceMile:
            name = 'mi'
            accuracy = 3
        elif default == DistanceNauticalMile:
            name = 'nm'
            accuracy = 3
        elif default == DistanceLine:
            name = 'ln'
            accuracy = 1
        elif default == DistanceMillimeter:
            name = 'mm'
            accuracy = 0
        elif default == DistanceCentimeter:
            name = 'cm'
            accuracy = 1
        elif default == DistanceMeter:
            name = 'm'
            accuracy = 2
        elif default == DistanceKilometer:
            name = 'km'
            accuracy = 3
        else:
            name = '?'
            accuracy = 6

        return f'{round(v, accuracy)} {name}'

    cpdef int units(self):
        return self._default_units



PressureMmHg = 40
PressureInHg = 41
PressureBar = 42
PressureHP = 43
PressurePSI = 44


cdef class Pressure:
    cdef double _value
    cdef int _default_units
    cdef __name__

    def __init__(self, value: double, units: int):
        self.__name__ = 'Pressure'
        self._value = self.to_default(value, units)
        self._default_units = units

    cdef double to_default(self, value: double, units: int):
            if units == PressureMmHg:
                return value
            elif units == PressureInHg:
                return value * 25.4
            elif units == PressureBar:
                return value * 750.061683
            elif units == PressureHP:
                return value * 750.061683 / 1000
            elif units == PressurePSI:
                return value * 51.714924102396
            else:
                raise KeyError(f'{self.__name__}: unit {units} is not supported')

    cdef double from_default(self, value: double, units: int):
        if units == PressureMmHg:
            return value
        elif units == PressureInHg:
            return value / 25.4
        elif units == PressureBar:
            return value / 750.061683
        elif units == PressureHP:
            return value / 750.061683 * 1000
        elif units == PressurePSI:
            return value / 51.714924102396
        else:
            raise KeyError(f'KeyError: {self.__name__}: unit {units} is not supported')

    cpdef double value(self, units: int):
        return self.from_default(self._value, units)

    cpdef Pressure convert(self, units: int):
        cdef double value = self.get_in(units)
        return Pressure(value, units)

    cpdef double get_in(self, units: int):
        return self.from_default(self._value, units)

    def __str__(self):
        return self.string()

    cdef string(self):
        cdef name
        cdef int accuracy
        cdef int default = self._default_units
        cdef double v = self.from_default(self._value, default)
        if default == PressureMmHg:
            name = 'mmHg'
            accuracy = 0
        elif default == PressureMmHg:
            name = 'inHg'
            accuracy = 2
        elif default == PressureBar:
            name = 'bar'
            accuracy = 2
        elif default == PressureHP:
            name = 'hPa'
            accuracy = 4
        elif default == PressurePSI:
            name = 'psi'
            accuracy = 4
        else:
            name = '?'
            accuracy = 6

        return f'{round(v, accuracy)} {name}'

    cpdef int units(self):
        return self._default_units



from libc.math cimport pi, atan, tan

AngularRadian = 0
AngularDegree = 1
AngularMOA = 2
AngularMil = 3
AngularMRad = 4
AngularThousand = 5
AngularInchesPer100Yd = 6
AngularCmPer100M = 7

cdef class Angular:
    cdef double _value
    cdef int _default_units
    cdef __name__

    def __init__(self, value: double, units: int):
        self.__name__ = 'Angular'
        self._value = self.to_default(value, units)
        self._default_units = units

    cdef double to_default(self, value: double, units: int):
        if units == AngularRadian:
            return value
        elif units == AngularDegree:
            return value / 180 * pi
        elif units == AngularMOA:
            return value / 180 * pi / 60
        elif units == AngularMil:
            return value / 3200 * pi
        elif units == AngularMRad:
            return value / 1000
        elif units == AngularThousand:
            return value / 3000 * pi
        elif units == AngularInchesPer100Yd:
            return atan(value / 3600)
        elif units == AngularCmPer100M:
            return atan(value / 10000)
        else:
            raise KeyError(f'{self.__name__}: unit {units} is not supported')

    cdef double from_default(self, value: double, units: int):
        if units == AngularRadian:
            return value
        elif units == AngularDegree:
            return value * 180 / pi
        elif units == AngularMOA:
            return value * 180 / pi * 60
        elif units == AngularMil:
            return value * 3200 / pi
        elif units == AngularMRad:
            return value * 1000
        elif units == AngularThousand:
            return value * 3000 / pi
        elif units == AngularInchesPer100Yd:
            return tan(value) * 3600
        elif units == AngularCmPer100M:
            return tan(value) * 10000
        else:
            raise KeyError(f'KeyError: {self.__name__}: unit {units} is not supported')

    cpdef double value(self, units: int):
        return self.from_default(self._value, units)

    cpdef Angular convert(self, units: int):
        cdef double value = self.get_in(units)
        return Angular(value, units)

    cpdef double get_in(self, units: int):
        return self.from_default(self._value, units)

    def __str__(self):
        return self.string()

    cdef string(self):
        cdef name
        cdef int accuracy
        cdef int default = self._default_units
        cdef double v = self.from_default(self._value, default)
        if default == AngularRadian:
            name = 'rad'
            accuracy = 6
        elif default == AngularDegree:
            name = '°'
            accuracy = 4
        elif default == AngularMOA:
            name = 'moa'
            accuracy = 2
        elif default == AngularMil:
            name = 'mil'
            accuracy = 2
        elif default == AngularMRad:
            name = 'mrad'
            accuracy = 2
        elif default == AngularThousand:
            name = 'ths'
            accuracy = 2
        elif default == AngularInchesPer100Yd:
            name = 'in/100yd'
            accuracy = 2
        elif default == AngularCmPer100M:
            name = 'cm/100m'
            accuracy = 2
        else:
            name = '?'
            accuracy = 6

        return f'{round(v, accuracy)} {name}'

    cpdef int units(self):
        return self._default_units
